#include <avr/io.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <util/delay.h>
#include "drm3310.h"
#include "message.h"

// Define baud rate
#define USART_BAUDRATE 19200
#define BAUD_PRESCALE (((F_CPU / (USART_BAUDRATE * 16UL))) - 1)

#define DEV_BUTTON

#ifdef DEV_BUTTON
#define SRC_ADR 1
#define DST_ADR 2
#endif /* DEV_BUTTON */

#ifdef DEV_SWITCH
#define SRC_ADR 2
#define DST_ADR 1
#endif /* DEV_SWITCH */

struct header_t {
	uint8_t version;
	uint8_t length;
	uint8_t source;
	uint8_t destination;
	uint8_t protocol;
	uint8_t reserved;
	uint8_t checksum;
};

void printHeader(struct header_t* header) {
	uint8_t str1[] = "Vers:   Len:  ";
	str1[5] = '0' + (header->version / 10) % 10;
	str1[6] = '0' + (header->version % 10);
	str1[12] = '0' + (header->length / 10) % 10;
	str1[13] = '0' + (header->length % 10);
	glcdn_PutStrXLine(str1, 0, 2);

	uint8_t str2[] = "Source    :   ";
	str2[11] = '0' + (header->source / 100);
	str2[12] = '0' + (header->source / 10) % 10;
	str2[13] = '0' + (header->source % 10);
	glcdn_PutStrXLine(str2, 0, 3);

	uint8_t str3[] = "Destinatio:   ";
	str3[11] = '0' + (header->destination / 100);
	str3[12] = '0' + (header->destination / 10) % 10;
	str3[13] = '0' + (header->destination % 10);
	glcdn_PutStrXLine(str3, 0, 4);

	uint8_t str4[] = "Prot:   Res:  ";
	str4[5] = '0' + (header->protocol / 10) % 10;
	str4[6] = '0' + (header->protocol % 10);
	str4[12] = '0' + (header->reserved / 10) % 10;
	str4[13] = '0' + (header->reserved % 10);
	glcdn_PutStrXLine(str4, 0, 5);

	uint8_t str5[] = "Checksum:     ";
	str5[11] = '0' + (header->checksum / 100);
	str5[12] = '0' + (header->checksum / 10) % 10;
	str5[13] = '0' + (header->checksum % 10);
	glcdn_PutStrXLine(str5, 0, 6);
}

void printMessage(struct message_t* message) {
	uint8_t str1[] = "Comp. Type: xx";
	str1[12] = '0' + (message->component_type / 10) % 10;
	str1[13] = '0' + (message->component_type % 10);
	glcdn_PutStrXLine(str1, 0, 2);

	uint8_t str2[] = "C. Sub Type: x";
	str2[13] = '0' + (message->component_subtype % 10);
	glcdn_PutStrXLine(str2, 0, 3);

	uint8_t str3[] = "Msg Type:   xx";
	str3[12] = '0' + (message->message_type / 10) % 10;
	str3[13] = '0' + (message->message_type % 10);
	glcdn_PutStrXLine(str3, 0, 4);

	uint8_t str4[] = "Message:     x";
	str4[13] = '0' + (message->message % 10);
	glcdn_PutStrXLine(str4, 0, 5);

	uint8_t str5[] = "Checksum:     ";
	str5[11] = '0' + (message->checksum / 100);
	str5[12] = '0' + (message->checksum / 10) % 10;
	str5[13] = '0' + (message->checksum % 10);
	glcdn_PutStrXLine(str5, 0, 6);
}

void usart_Init(void) {
	// Set baud rate
	UBRRH = (BAUD_PRESCALE >> 8); // Load upper 8-bits into the high byte of the UBRR register
	UBRRL = (BAUD_PRESCALE & 0xFF); // Load lower 8-bits into the low byte of the UBRR register

	UCSRC = (1<<URSEL)|(1<<UCSZ1)|(1<<UCSZ0);  // Asynchron 8N1

	// Enable receiver and transmitter and receive complete interrupt
	UCSRB |= (1<<RXEN)|(1<<TXEN);  // UART RX, TX
}

char usart_RecieveByte() {
	while(!(UCSRA & (1<<RXC)));
	char c = UDR;
	UCSRA |= (1<<RXC);
	return c;
}

void usart_SendByte(uint8_t c) {
    UCSRA |= (1<<TXC);           // TXC lschen
	while (!(UCSRA & (1<<UDRE)));
	UDR = c;
}

void usart_SendBlock(uint8_t *s, uint8_t len) {
	PORTD |= (1<<PD2);           // RS485-Tranciever vom C auf Senden schalten
	while (*s) {
		usart_SendByte(*s);
		s++;
	}
    while (!(UCSRA & (1<<TXC))); // Warte bis TXComplete
    PORTD &= ~(1<<PD2);          // RS485-Tranciever vom C wieder auf Empfang schalten
}

uint8_t recieveHeader(struct header_t* header) {
	uint8_t data[] = "     ";
	for(uint8_t i = 0; i < 5; i++) {
		data[i] = usart_RecieveByte();
	}

	header->version = data[0] & 0x0F;
	header->length = data[0] >> 4;
	header->source = data[1];
	header->destination = data[2];
	header->protocol = data[3] & 0x0F;
	header->reserved = data[3] >> 4;
	header->checksum = data[4];

	uint8_t checksum = data[0] + data[1] + data[2] + data[3];

	return header->checksum != checksum;
}

uint8_t recieveMessage(struct message_t* message) {
	struct header_t header;
	uint8_t error = recieveHeader(&header);

	if(!error) {
		uint8_t str[] = "Error (HCS)   ";
		glcdn_PutStrXLine(str, 0, 2);
		return error;
	}

	uint8_t data[] = "   ";
	for(uint8_t i = 0; i < 3; i++) {
		data[i] = usart_RecieveByte();
	}

	message->component_type = data[0] & 0x0F;
	message->component_subtype = (data[0] >> 3) & 0x7;
	message->component_nr = data[0] >> 7;
	message->message_type = data[1] & 0x0F;
	message->message = (data[1] >> 3) & 0x7;
	message->value_length = data[1] >> 7;
	message->checksum = data[2];
	uint8_t checksum = data[0] + data[1];

	printHeader(&header);

	return message->checksum != checksum;
}

void sendDatagram(uint8_t *message, uint8_t msgLen) {
	struct header_t header;
	header.version = 1;
	header.length = (msgLen+5+3)/4; // length*4 Byte
	header.source = SRC_ADR;
	header.destination = DST_ADR;
	header.protocol = 1;
	header.reserved = 0;
	header.checksum = 0;

	uint8_t data[] = "        ";
	data[0] = header.version + (header.length << 4);
	data[1] = header.source;
	data[2] = header.destination;
	data[3] = header.protocol + (header.reserved << 4);
	data[4] = data[0] + data[1] + data[2] + data[3];
	for(uint8_t i = 0; i < msgLen; i++) {
		data[5+i] = *(message+i);
	}
	usart_SendBlock(data, 8);

	printHeader(&header);
}

void sendMessage(struct message_t *message) {
	uint8_t multi_component = message->component_nr == 0 ? 0 : 1;
	uint8_t message_value = message->value_length == 0 ? 0 : 1;
	uint8_t data[] = "   ";
	data[0] = message->component_type + (message->component_subtype << 4) + (multi_component << 7);
	data[1] = message->message_type + (message->message << 4) + (message_value << 7);
	data[2] = data[0] + data[1];
	sendDatagram(data, 3);
}

#ifdef DEV_BUTTON
void main_loop(void) {
	uint8_t count = 0;
	while(1) {
		uint8_t str[] = "Sending:   :  ";
		str[12] = '0' + count/10;
		str[13] = '0' + count%10;
		glcdn_PutStrXLine(str, 0, 1);

		struct message_t message;
		message.component_type = MESSAGE_COMPONENT_TYPE_BUTTON;
		message.component_subtype = MESSAGE_COMPONENT_SUBTYPE_BUTTON_PUSH_BUTTON;
		message.component_nr = 0;
		message.message_type = MESSAGE_MESSAGE_TYPE_EVENT;
		message.message = MESSAGE_MESSAGE_BUTTON_PUSH_BUTTON_EVENT_PRESSED;
		message.value_length = 0;

		sendMessage(&message);

		_delay_ms(3000);
		printMessage(&message);

	    // turn LED on for 1 ms and turn it off again
		PORTD |= (1<<PD7);
		_delay_ms(1);
		PORTD &= ~(1 << PD7);
		_delay_ms(6999);
	}
}
#endif /* DEV_BUTTON */

#ifdef DEV_SWITCH
void main_loop(void) {
	uint8_t count = 0;
	while(1) {
		count = (count+1) % 100;
		uint8_t str[] = "Recieving:    ";
		str[12] = '0' + count/10;
		str[13] = '0' + count%10;
		glcdn_PutStrXLine(str, 0, 1);

		struct message_t message;
		if(recieveMessage(&message)) {
			uint8_t str[] = "Error (MCS)   ";
			glcdn_PutStrXLine(str, 0, 2);
			_delay_ms(3000);
		}
		else {
			_delay_ms(3000);
			printMessage(&message);
		}

		// turn LED on for 1 ms and turn it off again
		PORTD |= (1<<PD7);
		_delay_ms(1);
		PORTD &= ~(1 << PD7);
	}
}
#endif /* DEV_SWITCH */

int main(void) {
	_delay_ms(1000);

	glcdn_spiInit();
	glcdn_InitDisplay();
	glcdn_ClearLcd();

	glcdn_PutStrXLine((uint8_t*)"00:Init Max485", 0, 1);
	usart_Init();
	_delay_ms(1000);

	DDRD |= (1<<DDD2); // Datenrichtung D2 als Ausgang (MAX485)
	DDRD |= (1<<DDD7); // Datenrichtung D7 als Ausgang (LED)

	// Hauptschleife des Programms
	main_loop();
}
